// Build-script generated embedded data.
// The file included below is written by `build.rs` into `OUT_DIR`.
#![allow(dead_code)]
#![allow(clippy::all)]

include!(concat!(env!("OUT_DIR"), "/lipilekhika_generated_data.rs"));
